# Copyright 2014-2015 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""RPC helpers relating to DHCP."""

__all__ = [
    "configure",
    "DHCPv4Server",
    "DHCPv6Server",
]

import hashlib
from operator import itemgetter
import os

from provisioningserver.dhcp import (
    DHCPv4Server,
    DHCPv6Server,
)
from provisioningserver.dhcp.config import get_config
from provisioningserver.logger import get_maas_logger
from provisioningserver.rpc.exceptions import CannotConfigureDHCP
from provisioningserver.service_monitor import service_monitor
from provisioningserver.utils.fs import (
    read_text_file,
    sudo_delete_file,
    sudo_write_file,
)
from provisioningserver.utils.service_monitor import ServiceActionError
from provisioningserver.utils.shell import ExternalProcessError
from provisioningserver.utils.twisted import (
    asynchronous,
    synchronous,
)
from twisted.internet.defer import inlineCallbacks
from twisted.internet.threads import deferToThread


maaslog = get_maas_logger("dhcp")


@synchronous
def _write_config(server, failover_peers, shared_networks, hosts, interfaces):
    """Write the configuration file only if it has changed."""
    # Hash the old config and read the old interfaces file.
    old_config_hash = None
    old_interfaces_config = None
    if (os.path.exists(server.config_filename) and
            os.path.exists(server.interfaces_filename)):
        with open(server.config_filename, "rb") as fp:
            old_config = fp.read()
        old_config_hash = hashlib.md5(old_config).hexdigest()
        old_interfaces_config = read_text_file(
            server.interfaces_filename, encoding="utf-8")

    # Calculate the new config and new interfaces file.
    failover_peers = sorted(failover_peers, key=itemgetter("name"))
    shared_networks = sorted(shared_networks, key=itemgetter("name"))
    hosts = sorted(hosts, key=itemgetter("host"))
    new_dhcpd_config = get_config(
        server.template_basename, omapi_key=server.omapi_key,
        failover_peers=failover_peers, shared_networks=shared_networks,
        hosts=hosts).encode("utf-8")
    new_dhcpd_config_hash = hashlib.md5(new_dhcpd_config).hexdigest()
    new_interfaces = {
        interface['name']
        for interface in interfaces
    }
    new_interfaces_config = ' '.join(sorted(new_interfaces))

    # Check if the configuration really needs to be written.
    write_config = True
    if (old_config_hash is not None and
            old_config_hash == new_dhcpd_config_hash and
            old_interfaces_config is not None and
            old_interfaces_config == new_interfaces_config):
        write_config = False

    # Write the new config only if its different.
    if write_config:
        try:
            sudo_write_file(
                server.config_filename, new_dhcpd_config)
            sudo_write_file(
                server.interfaces_filename,
                new_interfaces_config.encode("utf-8"))
        except ExternalProcessError as e:
            # ExternalProcessError.__str__ contains a generic failure message
            # as well as the command and its error output. On the other hand,
            # ExternalProcessError.output_as_unicode contains just the error
            # output which is probably the best information on what went wrong.
            # Log the full error information, but keep the exception message
            # short and to the point.
            maaslog.error(
                "Could not rewrite %s server configuration (for network "
                "interfaces %s): %s", server.descriptive_name,
                new_interfaces_config, str(e))
            raise CannotConfigureDHCP(
                "Could not rewrite %s server configuration: %s" % (
                    server.descriptive_name, e.output_as_unicode))
    return new_interfaces_config, write_config


@synchronous
def _delete_config(server):
    """Delete the server config."""
    if os.path.exists(server.config_filename):
        sudo_delete_file(server.config_filename)


@asynchronous
@inlineCallbacks
def configure(server, failover_peers, shared_networks, hosts, interfaces):
    """Configure the DHCPv6/DHCPv4 server, and restart it as appropriate.

    :param server: A `DHCPServer` instance.
    :param failover_peers: List of dicts with failover parameters for each
        subnet where HA is enabled.
    :param shared_networks: List of dicts with shared network parameters that
        contain a list of subnets when the DHCP should server shared.
        If no shared network are defined, the DHCP server will be stopped.
    :param hosts: List of dicts with host parameters that
        contain a list of hosts the DHCP should statically.
    :param interfaces: List of interfaces that DHCP should use.
    """
    stopping = len(shared_networks) == 0

    if stopping:
        yield deferToThread(_delete_config, server)
        service = service_monitor.getServiceByName(server.dhcp_service)
        service.off()
        try:
            yield service_monitor.ensureService(server.dhcp_service)
        except ServiceActionError as e:
            # Error is already logged by the service monitor, nothing to
            # log for this exception.
            raise CannotConfigureDHCP(
                "%s server failed to stop: %s" % (
                    server.descriptive_name, e))
        except Exception as e:
            maaslog.error(
                "%s server failed to stop: %s",
                server.descriptive_name, e)
            raise CannotConfigureDHCP(
                "%s server failed to stop: %s" % (
                    server.descriptive_name, e))
    else:
        interfaces_config, needs_restart = yield deferToThread(
            _write_config, server,
            failover_peers, shared_networks, hosts, interfaces)
        service = service_monitor.getServiceByName(server.dhcp_service)
        service.on()
        if needs_restart:
            try:
                yield service_monitor.restartService(server.dhcp_service)
            except ServiceActionError as e:
                # Error is already logged by the service monitor, nothing to
                # log for this exception.
                raise CannotConfigureDHCP(
                    "%s server failed to restart: %s" % (
                        server.descriptive_name, e))
            except Exception as e:
                maaslog.error(
                    "%s server failed to restart (for network interfaces "
                    "%s): %s", server.descriptive_name, interfaces_config, e)
                raise CannotConfigureDHCP(
                    "%s server failed to restart: %s" % (
                        server.descriptive_name, e))
